import pytest

from api.models.offer import Offer
from api.models.stock import Stock
from api.models.user import User


class AccessorTest:
    def test_get_with_a_string_path(self, app):
        # Given
        offer = Offer(name='foo', type='bar')
        stock = Stock(offer=offer, price=1)

        # When
        offer_name = stock.get('offer.name')

        # Then
        assert offer_name == offer.name

    def test_get_with_a_index_path(self, app):
        # Given
        offer = Offer(name='foo', type='bar')
        stock = Stock(offer=offer, price=1)

        # When
        stock_price = stock.get('offer.stocks.0.price')

        # Then
        assert stock_price == stock.price



from functools import reduce
from itertools import groupby
from sqlalchemy import BigInteger, desc
from postgresql_audit.flask import versioning_manager

from sqlalchemy_api_handler.bases.accessor import Accessor
from sqlalchemy_api_handler.bases.errors import ActivityError
from sqlalchemy_api_handler.bases.save import Save
from sqlalchemy_api_handler.utils.datum import relationships_in
from sqlalchemy_api_handler.utils.logger import logger


def merged_datum_from_activities(activities,
                                 model,
                                 initial=None):
    return reduce(lambda agg, activity: {**agg, **relationships_in(activity.patch, model)},
                  activities,
                  relationships_in(initial, model) if initial else {})


class Activator(Save):

    @classmethod
    def get_activity(cls):
        return Activator.activity_cls

    @classmethod
    def set_activity(cls, activity_cls):
        Activator.activity_cls = activity_cls

    @staticmethod
    def activate(*activities,
                 with_check_not_soft_deleted=True):
        Activity = Activator.get_activity()
        for (entity_identifier, grouped_activities) in groupby(activities, key=lambda activity: activity.entityIdentifier):
            grouped_activities = sorted(grouped_activities, key=lambda activity: activity.dateCreated)

            first_activity = grouped_activities[0]
            table_name = first_activity.table_name
            model = Save.model_from_table_name(table_name)
            id_key = model.id.property.key

            if first_activity.verb == 'delete':
                query = model.query.filter_by(activityIdentifier=entity_identifier)
                entity = query.one()
                entity_id = entity.id
                query.delete()
                Activator.get_db().session.commit()
                delete_activity = entity.__deleteActivity__
                delete_activity.dateCreated = first_activity.dateCreated
                Save.save(delete_activity)

                # want to make as if first_activity was the delete_activity one
                # for such route like operations
                # '''
                #    ApiHandler.activate(**activities)
                #    return jsonify([as_dict(activity) for activity in activities])
                # '''
                first_activity.id = delete_activity.id
                if delete_activity.transaction:
                    first_activity.transaction = Activity.transaction.mapper.class_()
                    first_activity.transaction.actor  = delete_activity.transaction.actor
                Activator.activate(*grouped_activities[1:],
                                   with_check_not_soft_deleted=with_check_not_soft_deleted)
                continue

            if model is None:
                errors = ActivityError()
                errors.add_error('tableName', 'model from {} not found'.format(table_name))
                raise errors

            entity = model.query.filter_by(activityIdentifier=entity_identifier).first()
            entity_id = entity.id if entity else None
            if not entity_id:
                entity = model(**relationships_in(first_activity.patch, model))
                entity.activityIdentifier = entity_identifier
                Activator.save(entity)
                insert_activity = entity.__insertActivity__
                insert_activity.dateCreated = first_activity.dateCreated
                Save.save(insert_activity)
                # want to make as if first_activity was the insert_activity one
                # very useful for the routes operation
                # '''
                #    ApiHandler.activate(**activities)
                #    return jsonify([as_dict(activity) for activity in activities])
                # '''
                first_activity.id = insert_activity.id
                first_activity.changed_data = {**insert_activity.changed_data}
                if insert_activity.transaction:
                    first_activity.transaction = Activity.transaction.mapper.class_()
                    first_activity.transaction.actor  = insert_activity.transaction.actor
                Activator.activate(*grouped_activities[1:],
                                   with_check_not_soft_deleted=with_check_not_soft_deleted)
                continue

            min_date = min(map(lambda a: a.dateCreated, grouped_activities))
            already_activities_since_min_date = Activity.query \
                                                        .filter(
                                                            (entity._get_activity_join_filter()) & \
                                                            (Activity.dateCreated >= min_date)
                                                        ) \
                                                        .all()

            all_activities_since_min_date = sorted(already_activities_since_min_date + grouped_activities,
                                                   key=lambda activity: activity.dateCreated)

            entity = model.query.get(entity_id)
            before_data = all_activities_since_min_date[0].old_data \
                           or entity.just_before_activity_from(all_activities_since_min_date[0]).data
            merged_datum = {}
            for activity in all_activities_since_min_date:
                activity.old_data = before_data
                activity.verb = 'update'
                before_data = { **before_data,
                                 **activity.changed_data }
                merged_datum = { **merged_datum,
                                 **relationships_in(activity.patch, model) }

            if model.id.key in merged_datum:
                del merged_datum[model.id.key]

            db = Activator.get_db()
            db.session.add_all(grouped_activities)
            db.session.flush()
            entity.modify(merged_datum,
                          with_add=True,
                          with_check_not_soft_deleted=with_check_not_soft_deleted)
            db.session.flush()
            db.session.delete(entity.__lastActivity__)
            db.session.commit()


    @classmethod
    def models(cls):
        models = Accessor.models()
        Activity = cls.get_activity()
        if Activity:
            models += [Activity]
        return models


    def downgrade(op):
        op.drop_table('activity')
        op.drop_table('transaction')
        op.execute(
        '''
            DROP FUNCTION audit_table(regclass);
            DROP FUNCTION audit_table(regclass, text[]);
            DROP FUNCTION create_activity();
            DROP FUNCTION jsonb_subtract(jsonb, jsonb) CASCADE;
            DROP FUNCTION jsonb_change_key_name(jsonb, text, text);
        '''
        )

    def upgrade(op):
        from sqlalchemy_api_handler.mixins.activity_mixin import ActivityMixin
        db = Activator.get_db()
        versioning_manager.init(db.Model)
        versioning_manager.transaction_cls.__table__.create(op.get_bind())
        class Activity(ActivityMixin,
                       Activator,
                       versioning_manager.activity_cls):
            __table_args__ = {'extend_existing': True}

            id = versioning_manager.activity_cls.id
        Activity.__table__.create(op.get_bind())
        op.add_column('transaction',
                      sa.Column('actor_id',
                                sa.BigInteger()))



from datetime import datetime
import enum
from sqlalchemy import Column, \
                       Boolean, \
                       DateTime, \
                       Enum, \
                       String, \
                       Text
from sqlalchemy.dialects.postgresql import JSON, UUID
from sqlalchemy.sql import func


class TaskState(enum.Enum):
    CREATED = 'created'
    FAILED = 'failed'
    PUBLISHED = 'published'
    RECEIVED = 'received'
    RERUNNED = 'rerunned'
    STARTED = 'started'
    STOPPED = 'stopped'
    SUCCEED = 'succeed'


class TaskMixin(object):

    args = Column(JSON())

    celeryUuid = Column(UUID(as_uuid=True),
                             index=True,
                             nullable=False)

    creationTime = Column(DateTime(),
                          nullable=False,
                          server_default=func.now())

    hostname = Column(String(64))

    isEager = Column(Boolean())

    kwargs = Column(JSON())

    name = Column(String(256),
                  nullable=False)

    queue = Column(String(64))

    planificationTime = Column(DateTime())

    result = Column(JSON())

    state = Column(Enum(TaskState),
                   nullable=False)

    startTime = Column(DateTime())

    stopTime = Column(DateTime())

    traceback = Column(Text())



import sys
from datetime import datetime
from traceback import format_exception
import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSON, UUID
from sqlalchemy.sql import func

from sqlalchemy_api_handler.bases.activator import Activator
from sqlalchemy_api_handler.mixins.task_mixin import TaskState
from sqlalchemy_api_handler.utils.date import strptime


class Tasker(Activator):

    @classmethod
    def set_celery(cls, celery_app, flask_app):
        module = sys.modules['celery.signals']
        Task = Activator.model_from_name('Task')
        BaseTask = celery_app.Task
        db = Activator.get_db()
        task_db_session = sys.modules['flask_sqlalchemy'].SQLAlchemy().session

        class AppTask(BaseTask):
            abstract = True
            '''
            *_publish tasks run into APP
            task_* and after_return run into WORKER
            '''

            # see https://github.com/kwiersma/flask-celery-sqlalchemy/blob/master/app/celeryapp/__init__.py#L89
            def __call__(self, *args, **kwargs):
                if not celery_app.conf.CELERY_ALWAYS_EAGER:
                    with flask_app.app_context():
                        return BaseTask.__call__(self, *args, **kwargs)
                return BaseTask.__call__(self, *args, **kwargs)

            @module.before_task_publish.connect
            def create_task(body, headers, routing_key, **kwargs):
                celeryUuid = headers['id']
                task_entity = task_db_session.query(Task) \
                                             .filter_by(celeryUuid=celeryUuid) \
                                             .first()
                if task_entity:
                    task_entity.state = TaskState.RERUNNED
                else:
                    task_entity = Task(args=body[0],
                                       celeryUuid=headers['id'],
                                       creationTime=datetime.utcnow(),
                                       isEager=False,
                                       kwargs=body[1],
                                       name=headers['task'],
                                       planificationTime=strptime('{}Z'.format(headers['eta']).replace('+00:00', '')) \
                                                  if headers.get('eta') else None,
                                       queue=routing_key,
                                       state=TaskState.CREATED)
                task_db_session.add(task_entity)
                task_db_session.commit()

            @module.after_task_publish.connect
            def modify_task_to_published_state(headers, **kwargs):
                task_entity = task_db_session.query(Task) \
                                             .filter_by(celeryUuid=headers['id']) \
                                             .one()
                task_entity.state = TaskState.PUBLISHED
                task_db_session.add(task_entity)
                task_db_session.commit()

            @module.task_received.connect
            def modify_task_to_received_state(request, sender, **kwargs):
                task_entity = task_db_session.query(Task) \
                                             .filter_by(celeryUuid=request.id) \
                                             .first()
                if not task_entity:
                    task_entity = Task(celeryUuid=request.id,
                                       name=request.task.name)
                task_entity.hostname = sender.hostname
                task_entity.state = TaskState.RECEIVED
                task_db_session.add(task_entity)
                task_db_session.commit()

            @module.task_prerun.connect
            def modify_task_to_started_state(task_id, task, **kwargs):
                if task.request.is_eager:
                    task_entity = Task(celeryUuid=task_id,
                                       isEager=True,
                                       name=task.name)
                else:
                    task_entity = task_db_session.query(Task) \
                                                 .filter_by(celeryUuid=task_id) \
                                                 .one()
                task_entity.startTime = datetime.utcnow()
                task_entity.state = TaskState.STARTED
                task_db_session.add(task_entity)
                task_db_session.commit()

            @module.task_failure.connect
            def modify_task_to_failure_state(sender, traceback, **kwargs):
                task_entity = task_db_session.query(Task) \
                                             .filter_by(celeryUuid=sender.request.id) \
                                             .one()
                exc_type, exc_value, exc_traceback = sys.exc_info()
                task_entity.traceback = ' '.join(format_exception(exc_type,
                                                                  exc_value,
                                                                  traceback))
                task_entity.stopTime = datetime.utcnow()
                task_entity.state = TaskState.FAILED
                task_db_session.add(task_entity)
                task_db_session.commit()

            @module.task_success.connect
            def modify_task_to_success_state(sender, result, **kwargs):
                task_entity = task_db_session.query(Task) \
                                             .filter_by(celeryUuid=sender.request.id) \
                                             .one()
                task_entity.result = result
                task_entity.stopTime = datetime.utcnow()
                task_entity.state = TaskState.SUCCEED
                task_db_session.add(task_entity)
                task_db_session.commit()

            @module.task_postrun.connect
            def remove_session(*args, **kwargs):
                task_db_session.remove()

            @module.task_revoked.connect
            def modify_task_to_revoked_state(request, **kwargs):
                task_entity = task_db_session.query(Task) \
                                             .filter_by(celeryUuid=request.id) \
                                             .one()
                task_entity.stopTime = datetime.utcnow()
                task.state = TaskState.STOPPED
                task_db_session.add(task_entity)
                task_db_session.commit()
                task_db_session.remove()

            # see https://github.com/kwiersma/flask-celery-sqlalchemy/blob/master/app/celeryapp/__init__.py#L89
            def after_return(self, status, retval, task_id, args, kwargs, einfo):
                if not celery_app.conf.CELERY_ALWAYS_EAGER:
                    db.session.remove()

        celery_app.Task = AppTask

    def downgrade(op):
        op.drop_table('task')
        task_state = sa.Enum(name='taskstate')
        task_state.drop(op.get_bind())

    def upgrade(op):
        task_state = sa.Enum('CREATED',
                             'FAILED',
                             'PUBLISHED',
                             'RECEIVED',
                             'RERUNNED',
                             'STARTED',
                             'STOPPED',
                             'SUCCEED',
                              name='taskstate')
        op.create_table('task',
                        sa.Column('args', JSON()),
                        sa.Column('celeryUuid',
                                  UUID(as_uuid=True),
                                  index=True,
                                  nullable=False),
                        sa.Column('creationTime',
                                  sa.DateTime(),
                                  nullable=False,
                                  server_default=func.now()),
                        sa.Column('hostname',
                                  sa.String(64)),
                        sa.Column('isEager',
                                  sa.Boolean()),
                        sa.Column('kwargs',
                                  JSON()),
                        sa.Column('name',
                                  sa.String(256),
                                  nullable=False),
                        sa.Column('queue',
                                  sa.String(64)),
                        sa.Column('planificationTime',
                                  sa.DateTime()),
                        sa.Column('result',
                                  JSON()),
                        sa.Column('state',
                                  sa.Enum(TaskState),
                                  nullable=False),
                        sa.Column('startTime',
                                  sa.DateTime()),
                        sa.Column('stopTime',
                                  sa.DateTime()),
                        sa.Column('traceback',
                                  sa.Text()))
